/*
 * Copyright 2009-2010 Maarten Baert
 * maarten-baert@hotmail.com
 * http://www.maartenbaert.be/
 * 
 * This file is part of ExtremePhysics.
 * 
 * ExtremePhysics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * ExtremePhysics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with ExtremePhysics. If not, see <http://www.gnu.org/licenses/>.
 * 
 * File: ep_Contact_VelocityConstraints.h
 * Solves contact velocity constraints.
 * Included in ep_World_SolveVelocityConstraints.cpp.
 */

#include "ExtremePhysics.h"

EP_FORCEINLINE void ep_Contact::InitVelocityConstraints() {
	
	restitution = (double)(sqrt(shape1->restitution*shape2->restitution));
	friction = (double)(sqrt(shape1->friction*shape2->friction));
	
	for(int i = 0; i<2; ++i) {
		ep_ContactPoint *cp = contactpoints+i;
		if(cp->active) {
			
			// calculate normal mass
			{
				double m1 = cp->yy1*nx-cp->xx1*ny,
				       m2 = cp->yy2*nx-cp->xx2*ny;
				cp->normalmass = (1.0+world->mass_bias)/(
				  body1->invmass+body2->invmass
				  +ep_sqr(m1)*body1->invinertia
				  +ep_sqr(m2)*body2->invinertia);
			}
			
			// calculate tangent mass
			{
				double m1 = cp->yy1*ny+cp->xx1*nx,
				       m2 = cp->yy2*ny+cp->xx2*nx;
				cp->tangentmass = (1.0+world->mass_bias)/(
				  body1->invmass+body2->invmass
				  +ep_sqr(m1)*body1->invinertia
				  +ep_sqr(m2)*body2->invinertia);
			}
			
			// init target speeds
			cp->normaltargetvel = shape1->normalvelocity+shape2->normalvelocity;
			cp->tangenttargetvel = shape1->tangentvelocity+shape2->tangentvelocity;
			
			// baumgarte
			{
				double sep =
				  ((body1->x+cp->xx1)-(body2->x+cp->xx2))*nx+
				  ((body1->y+cp->yy1)-(body2->y+cp->yy2))*ny;
				// Use a small 'reversed' bias factor so forces can't build
				// up. Usually 'sep' is negative though.
				cp->normaltargetvel -= (sep<0.0)? world->baumgarte_factor*sep : world->baumgarte_factor*sep*0.01;
			}
			
			{
				// calculate current normal velocity
				double currnormalvel =
				   ((body1->xvel+cp->yy1*body1->rotvel)-(body2->xvel+cp->yy2*body2->rotvel))*nx
				  +((body1->yvel-cp->xx1*body1->rotvel)-(body2->yvel-cp->xx2*body2->rotvel))*ny;
				// restitution
				cp->normaltargetvel -= ep_min(0.0, currnormalvel+world->velocity_threshold)*restitution;
				// store speed error
				cp->normalveldelta = cp->normaltargetvel-currnormalvel;
			}
			
			{
				// calculate current tangent velocity
				double currtangentvel =
				   ((body1->xvel+cp->yy1*body1->rotvel)-(body2->xvel+cp->yy2*body2->rotvel))*ny
				  -((body1->yvel-cp->xx1*body1->rotvel)-(body2->yvel-cp->xx2*body2->rotvel))*nx;
				// store speed error
				cp->tangentveldelta = cp->tangenttargetvel-currtangentvel;
			}
			
		}
	}
	
}

EP_FORCEINLINE void ep_Contact::ApplyImpulses() {
	
	for(int i = 0; i<2; ++i) {
		ep_ContactPoint *cp = contactpoints+i;
		if(cp->active) {
			if(body1->storecontactimpulses && body2->storecontactimpulses) {
				
				// clamp stored impulses
				double d = friction*cp->normalforce;
				cp->tangentforce = ep_clamp(-d, d, cp->tangentforce);
				
				double dx = cp->normalforce*nx+cp->tangentforce*ny;
				double dy = cp->normalforce*ny-cp->tangentforce*nx;
				body1->_ApplyImpulse(cp->xx1, cp->yy1, +dx, +dy);
				body2->_ApplyImpulse(cp->xx2, cp->yy2, -dx, -dy);
				
			} else {
				cp->normalforce = 0.0;
				cp->tangentforce = 0.0;
			}
		}
	}
	
}

EP_FORCEINLINE void ep_Contact::SolveVelocityConstraints() {
	
	for(int i = 0; i<2; ++i) {
		ep_ContactPoint *cp = contactpoints+i;
		if(cp->active) {
			
			// normal contact constraints
			double currnormalvel =
			   ((body1->xvel+cp->yy1*body1->rotvel)-(body2->xvel+cp->yy2*body2->rotvel))*nx
			  +((body1->yvel-cp->xx1*body1->rotvel)-(body2->yvel-cp->xx2*body2->rotvel))*ny;
			double f = (cp->normaltargetvel-currnormalvel)*cp->normalmass;
			
			// f and normalforce are positive
			// always clamp the accumulated impulse
			if(f<-cp->normalforce) {
				f = -cp->normalforce;
				cp->normalforce = 0.0;
			} else {
				cp->normalforce += f;
			}
			
			if(f!=0.0) {
				double fx = f*nx,
				       fy = f*ny;
				body1->_ApplyImpulse(cp->xx1, cp->yy1, +fx, +fy);
				body2->_ApplyImpulse(cp->xx2, cp->yy2, -fx, -fy);
			}
			
		}
	}
	
	for(int i = 0; i<2; ++i) {
		ep_ContactPoint *cp = contactpoints+i;
		if(cp->active) {
			
			// tangent contact constraints (friction)
			double currtangentvel =
			   ((body1->xvel+cp->yy1*body1->rotvel)-(body2->xvel+cp->yy2*body2->rotvel))*ny
			  -((body1->yvel-cp->xx1*body1->rotvel)-(body2->yvel-cp->xx2*body2->rotvel))*nx;
			double f = (cp->tangenttargetvel-currtangentvel)*cp->tangentmass;
			
			// always clamp the accumulated impulse
			double bound = friction*cp->normalforce;
			if(cp->tangentforce+f<-bound) {
				f = -cp->tangentforce-bound;
				cp->tangentforce = -bound;
			} else if(cp->tangentforce+f>+bound) {
				f = -cp->tangentforce+bound;
				cp->tangentforce = +bound;
			} else {
				cp->tangentforce += f;
			}
			
			if(f!=0.0) {
				double fx = +f*ny,
				       fy = -f*nx;
				body1->_ApplyImpulse(cp->xx1, cp->yy1, +fx, +fy);
				body2->_ApplyImpulse(cp->xx2, cp->yy2, -fx, -fy);
			}
			
		}
	}
	
}

/*
Contact point derivation notes:

normal constraint: relative speed at contact point (#) <dot> contact normal is always zero

(# = relative coordinates of contact point)

fx = f*nx;
fy = f*ny;

delta-xspeed = fx/mass1+fx/mass2+( yy1#)*(fx*yy1-fy*xx1)/inertia1+( yy2#)*(fx*yy2-fy*xx2)/inertia2;
delta-yspeed = fy/mass1+fy/mass2+(-xx1#)*(fx*yy1-fy*xx1)/inertia1+(-xx2#)*(fx*yy2-fy*xx2)/inertia2;

delta-normalspeed = nx*delta-xspeed+ny*delta-yspeed;
delta-normalspeed =
  nx*(f*nx/mass1+f*nx/mass2+( yy1#)*(f*nx*yy1-f*ny*xx1)/inertia1+( yy2#)*(f*nx*yy2-f*ny*xx2)/inertia2)+
  ny*(f*ny/mass1+f*ny/mass2+(-xx1#)*(f*nx*yy1-f*ny*xx1)/inertia1+(-xx2#)*(f*nx*yy2-f*ny*xx2)/inertia2);
delta-normalspeed = f*(
  nx*nx/mass1+nx*nx/mass2+( yy1#)*nx*(nx*yy1-ny*xx1)/inertia1+( yy2#)*nx*(nx*yy2-ny*xx2)/inertia2)+
  ny*ny/mass1+ny*ny/mass2+(-xx1#)*ny*(nx*yy1-ny*xx1)/inertia1+(-xx2#)*ny*(nx*yy2-ny*xx2)/inertia2));

delta-normalspeed = f*(
  (nx*nx+ny*ny)/mass1+
  (nx*nx+ny*ny)/mass2+
  (yy1#*nx-xx1#*ny)*(yy1*nx-xx1*ny)/inertia1+
  (yy2#*nx-xx2#*ny)*(yy2*nx-xx2*ny)/inertia2);

nx*nx+ny*ny is always 1.0
if (xx1,yy1)==(xx1#,yy1#) and (xx2,yy2)==(xx2#,yy2#) then this formula can be simplified:

delta-normalspeed = f*(
(1.0)/mass1+
(1.0)/mass2+
sqr(yy1*nx-xx1*ny)/inertia1+
sqr(yy2*nx-xx2*ny)/inertia2);

tangent constraint: identical to normal constraint but nx is replaced by ny and ny is replaced by -nx
*/

/*
//TODO//
I'm not sure whether I will use this ...
if(c->contactpoints[0].active && c->contactpoints[1].active) { // use block solver
	
	// block solver
	// 
	// [ ([0],[0]) , ([0],[1]) ]
	// [ ([1],[0]) , ([1],[1]) ]
	
	// set up matrix
	aa =
	c->obj1->mass_factor+
	c->obj2->mass_factor+
	(c->contactpoints[0].yy1*c->nx-c->contactpoints[0].xx1*c->ny)*(c->contactpoints[0].yy1*c->nx-c->contactpoints[0].xx1*c->ny)*c->obj1->inertia_factor+
	(c->contactpoints[0].yy2*c->nx-c->contactpoints[0].xx2*c->ny)*(c->contactpoints[0].yy2*c->nx-c->contactpoints[0].xx2*c->ny)*c->obj2->inertia_factor;
	bb =
	c->obj1->mass_factor+
	c->obj2->mass_factor+
	(c->contactpoints[0].yy1*c->nx-c->contactpoints[0].xx1*c->ny)*(c->contactpoints[1].yy1*c->nx-c->contactpoints[1].xx1*c->ny)*c->obj1->inertia_factor+
	(c->contactpoints[0].yy2*c->nx-c->contactpoints[0].xx2*c->ny)*(c->contactpoints[1].yy2*c->nx-c->contactpoints[1].xx2*c->ny)*c->obj2->inertia_factor;
	cc =
	c->obj1->mass_factor+
	c->obj2->mass_factor+
	(c->contactpoints[1].yy1*c->nx-c->contactpoints[1].xx1*c->ny)*(c->contactpoints[0].yy1*c->nx-c->contactpoints[0].xx1*c->ny)*c->obj1->inertia_factor+
	(c->contactpoints[1].yy2*c->nx-c->contactpoints[1].xx2*c->ny)*(c->contactpoints[0].yy2*c->nx-c->contactpoints[0].xx2*c->ny)*c->obj2->inertia_factor;
	dd =
	c->obj1->mass_factor+
	c->obj2->mass_factor+
	(c->contactpoints[1].yy1*c->nx-c->contactpoints[1].xx1*c->ny)*(c->contactpoints[1].yy1*c->nx-c->contactpoints[1].xx1*c->ny)*c->obj1->inertia_factor+
	(c->contactpoints[1].yy2*c->nx-c->contactpoints[1].xx2*c->ny)*(c->contactpoints[1].yy2*c->nx-c->contactpoints[1].xx2*c->ny)*c->obj2->inertia_factor;
	
	// invert (mass matrix)
	d = 1.0/(aa*dd-bb*cc);
	c->col1_x = d*dd;
	c->col2_x = -d*bb;
	c->col1_y = -d*cc;
	c->col2_y = d*aa;
	
}
*/

